!negativeValuesInOut
! Module: asmooth_mod
!
! Authors:  Ian Stewart - SSC-LUX
!          
! Description: Performs adaptive smoothing of FITS images.
! 
! Comments:
!
!TODO:
!	- doco upgrade.
!	- check that, if min/maxsigma or ngauss are different to those used to make a template set, that the output still looks ok.
!	- Check accuracy of spher2ims for small (decA - decRef). (see equivalent rgslib routine)
!	- Do some careful conversion of real64 inputs.
!	- make sure test sets of other tasks fail if the task is not there! See GV mail of 29 May 2001. This is now done for asmooth.
!	- reinstitute the supercritical smoothing option?
!	- see if can reduce memory loading further by absorbing mask into indexImage.
!	- see if can speed up fft by sampling a single gaussian array instead of having to recalculate it all the time.
!	- 'samesnr' to 'sameamp' (see doco)
!	- change det mask to ccd mask??
!	- copy over keywords and dss stuff _before_ smoothing, because this is quicker than the smoothing but at the same time more likely to result in errors.
!	- Write more keywords to output images (to allow the user to identify the input dataset).
!	- Sort out how pixels for which smoothing (is undesired)/(has been unsuccesful) are indicated in indexImage and templateImage.
!	- Don't just cut away all the pad, just that which either extends beyond the original image bounds or is zero-valued.
!	- Check for 'hot pixels' (ie > ~10^6 times median value) after dividing by exposure map. These can throw out the FFT results (which are only single precision) if a weight image is not also used.
!	- Weed out obsolete TODOs!
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module asmooth_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  use param  
  use errorHandling

!use asmooth_common ! for convolverType declaration.

!use test_mod

  use array_utils
  use dal_aux
  use fftw_aux
  use save_image
  use ssc_misc	! for intToStr, fracVal.
  use sort_mod
use dump

  implicit none 

  private 

  ! Command-line parameters:
  logical ::&
    normalizePlease = .false.,&
    withUserWidths = .false.,&
    withWeightImage = .false.,&
    withOutMask = .false.,&
    writeBadMask = .false.,&
    writeTemplate = .false.,&
    withExpMap = .false.,&
    remultiplyByExpMap = .false.,&
    forceCalcType = .false.,&
    doCalcByFFT = .false.,&
    readVarianceSet = .false.,&
    writeVarianceSet = .false.,&
    writeConvolvers = .false.,&
    withIndexImage = .false.
  character(len=255) ::&
    inSetName = '',&
    weightImageName = '',&
    outMaskSetName = '',&
    badMaskSetName = '',&
    templateSetName = '',&
    expMapSetName = '',&
    outSetName = '',&
    tempImageSetName = '',&
    inVarianceSetName = '',&
    outVarianceSetName = '',&
    convolversSetName = '',&
    indexImageSetName = ''
  character(len=9) ::&
    smoothType = '',&
    widthDistribStyle = '',&
    expMapUseType = '',&
    kernelType = ''
  real(single) ::&
    minConvolverWidth = 0.0,&
    maxConvolverWidth = 0.0,&
    desiredSNR = 0.0,&
    convolverWidth = 0.0,&
    limitOnNumDirectOps = 0.0
  real(single), allocatable ::&
    userConvolverWidths(:)
  integer(int16) ::&
    numConvolvers = 0

  ! Other globals:
  integer(int16) ::&
    xStart,&
    yStart,&
    xFinis,&
    yFinis,&
    xsize,&
    ysize
  logical ::&
    forwardPlanExists = .false.,&
    reversePlanExists = .false.
  type(FftwPlanType) ::&
    fftForwardPlanNum, fftReversePlanNum
    ! type def is in fftw_aux.
  real(single), parameter ::&
    maxDynamicRange = 10.0**4,&
    minAllowedWeight = 0.0001 ! if weight (sum of normalized weight image
    ! times normalized convolver) gets lower than this, set the output to
    ! zero and set the pixel in the failure mask.
  integer(int32), parameter :: maxNumConvolvers = 999

logical(bool) :: testFlag = .false.
!integer(int16) :: txi = 46, tyi = 46, tci = 16
integer(int16) :: txi = 383, tyi = 268, tci = 16

  type gaussData
    real(single)   :: sigma, norm, snrImprovement
    integer(int32) :: numPixels
    integer(int16) :: halfWidthPixels
  end type gaussData
  type convolverType
    real(single), pointer :: array(:,:)
  end type convolverType

  public asmooth

contains 
  
  !---------------------------------------------------------------------
  subroutine asmooth
  !---------------------------------------------------------------------

    real(single), pointer ::&
      inImage(:,:), expMapImage(:,:), weightImage(:,:), inVarImage(:,:),&
      weightedInImage(:,:), weightedVarImage(:,:)
    logical(bool), pointer ::&
      outMask(:,:)
    logical(bool), allocatable ::&
      failureMask(:,:), failureMaskOrigSize(:,:)
    real(single), allocatable ::&
      smoothedImage(:,:), smoothedImageOrigSize(:,:),&
      outVarImage(:,:), outVarImageOrigSize(:,:)
    integer(int32), pointer :: indexImage(:,:)
    integer(int16) :: halfWidthPixels, convolverXSize, convolverYSize
    integer(int32) :: i
    type(convolverType), pointer :: convolvers(:)
    type(gaussData), pointer :: gaussians(:)
    character(8) :: numBadPixStr
real(single) :: medianPositive

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    !....................................................................
    ! Load the input image to be smoothed and get the required dimensions
    ! of all other input images from that:

    call readArrayData(inSetName, inImage)
    ! in ssclib/dal_aux.
    xsize = size(inImage, 1)	! global
    ysize = size(inImage, 2)	! global

    !....................................................................
    if (withOutMask) then
      call readArrayData(outMaskSetName, outMask) ! may use the
      ! untreated exposure map as the outMask. Routine is in ssclib/dal_aux.

      if (size(outMask, 1) .ne. xsize .or. size(outMask, 2) .ne. ysize)&
call error('outMaskWrongSize',"The dimensions of --outmaskset are different &
&to those of --inset.")
    else
      allocate(outMask(xsize, ysize))
      outMask = .true.
    end if

    call maskArrayLimits(outMask, xStart, xFinis, yStart, yFinis)
    ! These limits are global variables. The routine is in ssclib/array_utils.

    call selectArraySubset(outMask, xStart, xFinis, yStart, yFinis)
    ! in ssclib/array_utils.

    if (count(outMask) <= 0)&
call error('wholeImageUnmasked', "The number of outMask pixels is 0. &
&Can't smooth masked image.")

    call selectArraySubset(inImage, xStart, xFinis, yStart, yFinis)
    ! in ssclib/array_utils.

    !....................................................................
    if (withWeightImage) then
      call readArrayData(weightImageName, weightImage)
      ! in ssclib/dal_aux.

      if (size(weightImage, 1) .ne. xsize .or. size(weightImage, 2) .ne. ysize)&
call error('weightImageWrongSize',"The dimensions of --weightset are &
&different to those of --inset.")

      call selectArraySubset(weightImage, xStart, xFinis, yStart, yFinis)
      ! in ssclib/array_utils.

    else ! make a dummy weight image:
      allocate(weightImage(xStart:xFinis, yStart:yFinis))
      weightImage = 1.0 
    end if

    weightImage = weightImage / maxval(weightImage)
    ! We need to do this because of later test against minAllowedWeight.

    !....................................................................
    if (withExpMap) then
      call readArrayData(expMapSetName, expMapImage)
      ! in ssclib/dal_aux.

if (testFlag) print *, 'expMapImage(txi,tyi):', expMapImage(txi,tyi)
      if (size(expMapImage, 1) .ne. xsize&
      .or. size(expMapImage, 2) .ne. ysize)&
call error('exposureMapWrongSize',"The dimensions of --expimageset are &
&different to those of --inset.")

      call selectArraySubset(expMapImage, xStart, xFinis, yStart, yFinis)
      !  in ssclib/array_utils.
      where(expMapImage==0.0) weightImage = 0.0
    end if

if (any(weightImage < 0.0)) call error('weightImageNegativeValues',"Some of &
&the pixels of --weightset have negative values.")

if (.not. any(weightImage > 0.0)) call error('weightImageAllZero'&
,"All the pixels of --weightset are zero-valued.")

    !....................................................................
    ! Set up the input variance image:

    if (.not. readVarianceSet .and. smoothType .ne. 'adaptive') then
    ! make a dummy variance image:

      allocate(inVarImage(xStart:xFinis, yStart:yFinis))
      inVarImage = 1.0

    else
      if (readVarianceSet) then
        call readArrayData(inVarianceSetName, inVarImage)
        ! in ssclib/dal_aux.

        if (size(inVarImage, 1) .ne. xsize&
        .or. size(inVarImage, 2) .ne. ysize)&
call error('varianceImageWrongSize',"The dimensions of --invarianceset are &
&different to those of --inset.")

        call selectArraySubset(inVarImage, xStart, xFinis, yStart, yFinis)
        ! in ssclib/array_utils.

if (any(inVarImage < 0.0)) call error('varianceImageNegativeValues'&
,"Some of the pixels of --invarianceset have negative values.")

      else ! logic dictates that smoothType must be 'adaptive' at this
      ! point in the IF statement. In this case a variance image is
      ! definitely needed, and, since it has not been supplied by the user,
      ! the only other thing that can be done is to assume the input image
      ! is Poissonian, and take this as its own variance estimator.

        if (any(inImage < 0.0))&
call error('inImageNegativeValues', 'The image contains some values less than &
&zero. Because of this it is not possible to calculate a variance image. &
&You should supply one via the appropriate task parameters.')

!*** other tests to see if inImage is poissonian?
        allocate(inVarImage(xStart:xFinis, yStart:yFinis))

        inVarImage = inImage
      end if
    end if

    !....................................................................
    ! Main routine:

    select case(smoothType)
    case('simple')
      numConvolvers = 1
      allocate(indexImage(xStart:xFinis, yStart:yFinis))
      indexImage = 1

      select case(kernelType)
      case('gaussian')

        halfWidthPixels = convolverArrayHalfWidth(convolverWidth)
        convolverXSize = 3 + 2 * halfWidthPixels
        convolverYSize = 3 + 2 * halfWidthPixels
        allocate(convolvers(1))
        allocate(convolvers(1)%array(convolverXSize, convolverYSize))
        call calcGaussianConvolver(convolverWidth, convolvers(1)%array)

      case('tophat')
        halfWidthPixels = floor(convolverWidth)
        convolverXSize = 1 + 2 * halfWidthPixels
        convolverYSize = 1 + 2 * halfWidthPixels
        allocate(convolvers(1))
        allocate(convolvers(1)%array(convolverXSize, convolverYSize))
        call calcTopHatConvolver(convolverWidth, convolvers(1)%array)

      case('squarebox')
        halfWidthPixels = floor(convolverWidth)
        convolverXSize = 1 + 2 * halfWidthPixels
        convolverYSize = 1 + 2 * halfWidthPixels
        allocate(convolvers(1))
        allocate(convolvers(1)%array(convolverXSize, convolverYSize))
        convolvers(1)%array = 1.0

      case default
call error('badConvolverStyle',"The value of parameter --convolverstyle was not recognized.")
      end select

    case('template')
      call calcConvolversFromTemplate(indexImage, convolvers)
!call dumpImageToFits(indexImage, 'harry_index.ds')

    case('adaptive')
      allocate(weightedInImage(xStart:xFinis, yStart:yFinis))
      allocate(weightedVarImage(xStart:xFinis, yStart:yFinis))

      if (withExpMap) then
        weightedInImage = divideImages(inImage * weightImage, expMapImage)
        ! leaves zeros in weightedInImage at pixels where expMapImage = 0.

        if (expMapUseType .eq. 'samesnr') then
           weightedVarImage = divideImages(inVarImage * weightImage&
             * weightImage, expMapImage * expMapImage)
           ! leaves zeros in weightedVarImage at pixels where expMapImage = 0.

        else if (expMapUseType .eq. 'samesize') then
          weightedVarImage = divideImages(inVarImage * weightImage&
            * weightImage, expMapImage * maxval(expMapImage))
          ! leaves zeros in weightedVarImage at pixels where expMapImage = 0.
        else
call error('badExpMapUse', 'The value '//trim(expMapUseType)//' of &
&--expmapuse is not recognised.')
        end if

      else
        weightedInImage  = inImage    * weightImage
        weightedVarImage = inVarImage * weightImage * weightImage
      end if

      medianPositive = median(abs(weightedInImage)&
        , logical(weightedInImage.ne.0.0, bool))

      call dynamicRangeWarnings(medianPositive, maxval(abs(weightedInImage)))

      call calcConvolversAdaptive(weightedInImage, weightedVarImage&
       , weightImage, outMask, indexImage, convolvers, gaussians)

!call dumpImageToFits(indexImage, 'fred_index.ds')
      deallocate(weightedInImage, weightedVarImage)

    case('withset')
      call readIndexAndConvolvers(indexImage, convolvers)
!call dumpImageToFits(indexImage, 'bert_index.ds')

    case default
call error('badSmoothStyle',"The value of parameter --smoothstyle was &
&not recognized.")
    end select

    !....................................................................
    !....................................................................
    ! Now the actual smoothing:

    allocate(smoothedImage(xStart:xFinis, yStart:yFinis))
    allocate(outVarImage(xStart:xFinis, yStart:yFinis))
    allocate(failureMask(xStart:xFinis, yStart:yFinis))

    if (withExpMap) then
      inImage = divideImages(inImage * weightImage, expMapImage)
      ! leaves zeros in inImage at pixels where expMapImage = 0.

      inVarImage = divideImages(inVarImage * weightImage&
        * weightImage, expMapImage * expMapImage)
      ! leaves zeros in inVarImage at pixels where expMapImage = 0.
    else
      inImage  = inImage    * weightImage
      inVarImage = inVarImage * weightImage * weightImage
    end if

    medianPositive = median(abs(inImage), logical(inImage.ne.0.0, bool))
    call dynamicRangeWarnings(medianPositive, maxval(abs(inImage)))

    call doSmoothing(inImage, inVarImage, weightImage, outMask, indexImage&
      , convolvers, smoothedImage, outVarImage, failureMask)

    !....................................................................
    ! Clean up fftw plans:

    if (forwardPlanExists) then
      call fftwnd_f77_destroy_plan(fftForwardPlanNum)
      forwardPlanExists = .false.
    end if
    if (reversePlanExists) then
      call fftwnd_f77_destroy_plan(fftReversePlanNum)
      reversePlanExists = .false.
    end if

    !....................................................................
    ! Image array outputs:

    if (writeVarianceSet) then
      if (.not.(readVarianceSet .or. smoothType .eq. 'adaptive'))&
        call warning('uselessVarianceImage',"No sensible input variance &
          &image was generated or supplied, so the --outvarianceset &
          &output is meaningless.")
!**** not yet in docs.

      call message(AppMsg, Verbose, 'Writing output variance image '&
      //trim(outVarianceSetName)//'.')

      where(failureMask) outVarImage = 0.0
      if (withExpMap) then
        if (remultiplyByExpMap) outVarImage = outVarImage * expMapImage&
          * expMapImage
      end if

      allocate(outVarImageOrigSize(xsize, ysize))
      outVarImageOrigSize = 0.0
      outVarImageOrigSize(xStart:xFinis, yStart:yFinis) = outVarImage
      call saveImage(outVarImageOrigSize, inSetName, outVarianceSetName&
        , tempImageSetName, xStart, yStart)
      deallocate(outVarImageOrigSize, outVarImage)
    end if

    deallocate(inImage, inVarImage, weightImage, outMask)

    if (any(failureMask)) then
      numBadPixStr = intToStr(count(failureMask))
      write(errstr, '(a, f9.7, a)') 'The sum over weights was found to &
        &be smaller than the limit of ', minAllowedWeight&
        , ' at '//trim(numBadPixStr)//' pixels. The output has been set &
        &to zero at these pixels. A map of these pixels may be obtained &
        &by setting --writebadmaskset=yes.'

      call warning('outMaskTooNarrow')

      where(failureMask) smoothedImage = 0.0
    end if

    if (writeBadMask) then
      call message(AppMsg, Verbose, 'Writing bad pixel mask '&
      //trim(badMaskSetName)//'.')

      allocate(failureMaskOrigSize(xsize, ysize))
      failureMaskOrigSize = .false.
      failureMaskOrigSize(xStart:xFinis, yStart:yFinis) = failureMask
      call saveImage(failureMaskOrigSize, inSetName, badMaskSetName&
        , tempImageSetName, xStart, yStart)
      deallocate(failureMaskOrigSize, failureMask)
    end if

    if (smoothType .eq. 'adaptive') then
      if (writeTemplate) call calcAndSaveTemplate(gaussians, indexImage)
      deallocate(gaussians)
!*** can get rid of gaussians at this level of the code if templates are dispensed with.

      if (writeConvolvers) call writeIndexAndConvolvers(indexImage, convolvers)
    end if

    do i = 1, numConvolvers
      deallocate(convolvers(i)%array)
    end do
    deallocate(indexImage, convolvers)

    if (withExpMap) then
      if (remultiplyByExpMap) smoothedImage = smoothedImage * expMapImage
      deallocate(expMapImage)
    end if

    call message(AppMsg, Verbose, 'Writing smoothed output image '&
    //trim(outSetName)//'.')

    allocate(smoothedImageOrigSize(xsize, ysize))
    smoothedImageOrigSize = 0.0
    smoothedImageOrigSize(xStart:xFinis, yStart:yFinis) = smoothedImage
    call saveImage(smoothedImageOrigSize, inSetName, outSetName&
      , tempImageSetName)

!*** write convolver info (+ name of orig file) to output.

    deallocate(smoothedImageOrigSize, smoothedImage)

  end subroutine asmooth

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    inSetName = stringParameter('inset')
    outSetName = stringParameter('outset')
    tempImageSetName = stringParameter('tempset')

    smoothType = stringParameter('smoothstyle')
    select case(smoothType)
    case('simple')
      kernelType = stringParameter('convolverstyle')
      convolverWidth = realParameter('width')
      normalizePlease = booleanParameter('normalize')

    case('adaptive')
      withUserWidths = booleanParameter('withuserwidths')
      if (withUserWidths) then
        numConvolvers = parameterCount('userwidths')
        allocate(userConvolverWidths(numConvolvers))
        do i = 1, numConvolvers
          userConvolverWidths(i) = realParameter('userwidths', i-1)
        end do

        if (any(userConvolverWidths(:numConvolvers-1)&
          >= userConvolverWidths(2:))) then
call error('badUserWidths', "The list --userwidths must be &
&monotonically increasing.")
        end if

      else
        numConvolvers = intParameter('nconvolvers')
        minConvolverWidth = realParameter('minwidth')
        maxConvolverWidth = realParameter('maxwidth')
        widthDistribStyle = stringParameter('widthliststyle')

        if (minConvolverWidth >= maxConvolverWidth) then
call error('badWidthRange', "The value of --maxwidth is less than or &
&equal to --minwidth.")
        end if
      end if

      desiredSNR = realParameter('desiredsnr')
      writeTemplate = booleanParameter('writetemplateset')
      if (writeTemplate) templateSetName = stringParameter('outtemplateset')
      writeConvolvers = booleanParameter('writeconvolvers')
      if (writeConvolvers) then
        convolversSetName = stringParameter('outconvolversset')
        indexImageSetName = stringParameter('outindeximageset')
      end if

      normalizePlease = .true.

    case('withset')
      convolversSetName = stringParameter('inconvolversset')
      withIndexImage = booleanParameter('withindeximageset')
      if (withIndexImage) then
        indexImageSetName = stringParameter('inindeximageset')
      end if
      normalizePlease = booleanParameter('normalizeset')

    case('template')
      templateSetName = stringParameter('intemplateset')

      normalizePlease = .true.

    case default
call error('badSmoothStyle',"The value of parameter --smoothstyle was &
&not recognized.")
    end select

    limitOnNumDirectOps = realParameter('nopslimit')
    forceCalcType = booleanParameter('forcecalctype')
    if (forceCalcType) doCalcByFFT = booleanParameter('calcbyfft')

    readVarianceSet = booleanParameter('readvarianceset')
    if (readVarianceSet) inVarianceSetName = stringParameter('invarianceset')
    writeVarianceSet = booleanParameter('writevarianceset')
    if (writeVarianceSet) outVarianceSetName = stringParameter('outvarianceset')

    withWeightImage = booleanParameter('withweightset')
    if (withWeightImage) weightImageName = stringParameter('weightset')
    withOutMask = booleanParameter('withoutmaskset')
    if (withOutMask) outMaskSetName = stringParameter('outmaskset')
    writeBadMask = booleanParameter('writebadmaskset')
    if (writeBadMask) badMaskSetName = stringParameter('badmaskset')

    withExpMap = booleanParameter('withexpimageset')
    if (withExpMap) then
      expMapSetName = stringParameter('expimageset')
      expMapUseType = stringParameter('expmapuse')
      remultiplyByExpMap = booleanParameter('remultiply')
    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
  function convolverArrayHalfWidth(gaussianSigma)
  !---------------------------------------------------------------------
  ! This is a function because it is called in a couple of places, and we
  ! may want to change the factor of 5.0 some day.

    implicit none

    real(kind=single), intent(in) ::&
      gaussianSigma
    integer(kind=int16) :: convolverArrayHalfWidth

    !-------------------------------------------------------------------
    ! Start:

    convolverArrayHalfWidth = floor(5.0 * gaussianSigma)

  end function convolverArrayHalfWidth

  !---------------------------------------------------------------------
  subroutine calcGaussianConvolver(sigma, convolver)
  !---------------------------------------------------------------------
    
    implicit none

    real(single), intent(in)  :: sigma
    real(single), intent(out) :: convolver(:,:)

    logical, save :: firstCall = .true.
    real(double), save ::&
      sigma22
    integer(int16) ::&
      convolverXSize,&
      convolverYSize,&
      convolverCentreX, convolverCentreY, i
    integer(int16), allocatable :: indices(:)
    real(double) :: xValues(size(convolver,1)), yValues(size(convolver,2))

    !-------------------------------------------------------------------
    ! Start:

    if (firstCall) sigma22 = 2.0 * sigma * sigma

    convolverXSize = size(convolver, 1)
    convolverYSize = size(convolver, 2)
    convolverCentreX = ceiling(convolverXSize / 2.0)
    convolverCentreY = ceiling(convolverYSize / 2.0)

    if (sigma22 > 0.0) then
      allocate(indices(convolverXSize))
      indices = (/(i-convolverCentreX, i=1,convolverXSize)/)
      xValues = exp(-(indices * indices) / sigma22)
      deallocate(indices)

      allocate(indices(convolverYSize))
      indices = (/(i-convolverCentreY, i=1,convolverYSize)/)
      yValues = exp(-(indices * indices) / sigma22)
      deallocate(indices)

      convolver = spread(xValues, 2, convolverYSize)&
        * spread(yValues, 1, convolverXSize)

    else
      convolver = 0.0
      convolver(convolverCentreX, convolverCentreY) = 1.0
    end if

  end subroutine calcGaussianConvolver

  !---------------------------------------------------------------------
  subroutine calcTopHatConvolver(radius, convolver)
  !---------------------------------------------------------------------
    
    implicit none

    real(single), intent(in)  :: radius
    real(single), intent(out) :: convolver(:,:)

    real(single) :: radiusSquared, x, y
    integer(int16) ::&
      convolverXSize,&
      convolverYSize,&
      convolverCentreX, convolverCentreY, xi, yi

    !-------------------------------------------------------------------
    ! Start:

    radiusSquared = radius * radius

    convolverXSize = size(convolver, 1)
    convolverYSize = size(convolver, 2)
    convolverCentreX = ceiling(convolverXSize / 2.0)
    convolverCentreY = ceiling(convolverYSize / 2.0)

    convolver = 0.0 ! default
    do xi = 1, convolverXSize
      x = xi - convolverCentreX
      do yi = 1, convolverYSize
        y = yi - convolverCentreY
        if (x * x + y * y <= radiusSquared) convolver(xi,yi) = 1.0
      end do
    end do

  end subroutine calcTopHatConvolver

  !---------------------------------------------------------------------
  subroutine calcConvolversFromTemplate(indexImage, convolvers)
  !---------------------------------------------------------------------
!*** NOTE that pre-2.7 templates are no longer supported.

    implicit none

    !-------------------------------------------------------------------
    ! Start

    integer(int32), pointer :: indexImage(:,:)
    type(convolverType), pointer :: convolvers(:)

    real(single), pointer :: templateImage(:,:)
    integer(int32) :: vSize, i
    real(single), allocatable :: tempVector(:)
    real(single) :: convolverWidths(maxNumConvolvers), currentWidth
    type(gaussData), pointer :: gaussians(:)

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Reading template image '&
      //trim(templateSetName)//'.')

    call readArrayData(templateSetName, templateImage)
    ! in ssclib/dal_aux.
    call selectArraySubset(templateImage, xStart, xFinis, yStart, yFinis)
    ! in ssclib/array_utils.
    allocate(indexImage(xStart:xFinis, yStart:yFinis))
    indexImage = 0

    vSize = size(templateImage, 1) * size(templateImage, 2)
    allocate(tempVector(vSize))
    tempVector = reshape(templateImage, (/vSize/))
    call quickSort(tempVector) ! in ssclib/sort_mod.

    numConvolvers = 0
    i = vSize
    do while(i > 1)
      currentWidth = tempVector(i)
!*** exit if currentWidth < 0? Indicates 'don't smooth here' so indexImage should remain at 0??
      numConvolvers = numConvolvers + 1
if (numConvolvers > maxNumConvolvers) call error('tooManyConvolvers'&
,"The number of convolvers calculated from the template image has exceeded &
&the limiting value.")
      convolverWidths(numConvolvers) = currentWidth

      where(templateImage == currentWidth) indexImage = numConvolvers

      i = smallestTrueIndex(logical(tempVector==currentWidth, bool)) - 1
    end do

    call calcGaussianLibrary(convolverWidths(:numConvolvers)&
      , gaussians, convolvers)

    deallocate(templateImage, gaussians)

  end subroutine calcConvolversFromTemplate

  !---------------------------------------------------------------------
  subroutine dynamicRangeWarnings(medianAbsValue, maxImageAbsValue)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: medianAbsValue, maxImageAbsValue

    real(single) :: dynamicRange

    !-------------------------------------------------------------------
    ! Start

!*** not yet included in the doco.

!print *, 'medianPositive, maxval(abs(weightedInImage)):', medianAbsValue, maxImageAbsValue

  if (medianAbsValue <= 0.0) return

    dynamicRange = maxImageAbsValue/medianAbsValue
    if (dynamicRange > maxDynamicRange&
      .and. (.not. forceCalcType .or. doCalcByFFT)) then

      if (withExpMap) then
        if (withWeightImage) then
          write(errstr, '(a, es10.4, a, es10.4, a)') "The ratio ", dynamicRange&
           , " between the largest and median values of the weighted, &
           &exposure-corrected image is larger than the warning limit of "&
           , maxDynamicRange, ". This will severely reduce the accuracy &
           &of any Fourier calculations. You should consider using an &
           &exposure map which better matches the input image, or &
           &removing the extreme-valued pixels from either --weightset &
           &or --outmaskset, or forcing the convolution to be non-Fourier &
           &(even though this can take much longer to process)."

        else
          write(errstr, '(a, es10.4, a, es10.4, a)') "The ratio ", dynamicRange&
           , " between the largest and median values of the exposure-&
           &corrected image is larger than the warning limit of "&
           , maxDynamicRange, ". This will severely reduce the accuracy &
           &of any Fourier calculations. You should consider also using &
           &a weight image, or using an exposure map which better matches &
           &the input image, or removing the extreme-valued pixels from &
           &--outmaskset, or forcing the convolution to be non-Fourier &
           &(even though this can take much longer to process)."
        end if

      else
        if (withWeightImage) then
          write(errstr, '(a, es10.4, a, es10.4, a)') "The ratio ", dynamicRange&
           , " between the largest and median values of the weighted &
           &image is larger than the warning limit of ", maxDynamicRange&
           , ". This will severely reduce the accuracy of any Fourier &
           &calculations. You should consider using an exposure map &
           &which better matches the input image, or removing the &
           &extreme-valued pixels from either --weightset or &
           &--outmaskset, or forcing the convolution to be non-Fourier &
           &(even though this can take much longer to process)."

        else
          write(errstr, '(a, es10.4, a, es10.4, a)') "The ratio ", dynamicRange&
           , " between the largest and median values of the image is &
           &larger than the warning limit of ", maxDynamicRange, ". This &
           &will severely reduce the accuracy of any Fourier calculations. &
           &You should consider also using a weight image, or using &
           &an exposure map which better matches the input image, or &
           &removing the extreme-valued pixels from --outmaskset, or &
           &forcing the convolution to be non-Fourier (even though &
           &this can take much longer to process)."
        end if
      end if

      call warning('dynamicRangeTooLarge')
    end if

  end subroutine dynamicRangeWarnings

  !---------------------------------------------------------------------
  subroutine calcConvolversAdaptive(inImage, inVarImage, weightImage, outMask&
  , indexImageArg, convolvers, gaussians)
  !---------------------------------------------------------------------
  ! Global variables:
  !	xStart, xFinis, yStart, yFinis: record the borders of the cropped
  ! input image.
  !	All parameter-set valriables naturally, including
  !		readVarianceSet
  !		withExpMap
  !		expMapUseType
  !		numConvolvers
  !		forceCalcType
  !		doCalcByFFT
  !		limitOnNumDirectOps
  !
  !	Globally-declared type gaussData is accessed.
!*** thorough check of all listings of globals needed...

    implicit none

    real(single),        intent(in) :: inImage(:,:),&
                                       inVarImage(size(inImage,1)&
                                                 ,size(inImage,2)),&
                                       weightImage(size(inImage,1)&
                                                  ,size(inImage,2))
    logical(bool),       intent(in) :: outMask(size(inImage,1)&
                                              ,size(inImage,2))
    integer(int32),      pointer    :: indexImageArg(:,:)
    type(convolverType), pointer    :: convolvers(:)
    type(gaussData),     pointer    :: gaussians(:)

    integer(int8), allocatable :: indexImage(:,:)
    integer(int16) ::&
      localXSize, localYSize,&
      convolverXSize, convolverYSize,&
      maxConvolverXSize, maxConvolverYSize,&
      paddedXSize, paddedYSize
    integer(int32) :: numPixels, numMaskedPixels
    real(single) ::&
      sigma
    integer(int8) :: i
    complex(single), allocatable ::&
      inImageFFT(:,:), inVarImageFFT(:,:), inWeightFFT(:,:)
    real(single), allocatable ::&
      smoothedImage(:,:), bestSnr(:,:), secondBestSnr(:,:),&
      smoothedImageUncerts(:,:), smoothedImageVariance(:,:)
    type(convolverType), pointer :: convsSqurd(:)
    logical(bool), allocatable :: failureMask(:,:)
    real(single), pointer :: convolverWidths(:)

    !-------------------------------------------------------------------
    ! Start:

    localXSize = 1 + xFinis - xStart
    localYSize = 1 + yFinis - yStart
    numMaskedPixels = count(outMask)

    !....................................................................
    ! Set up the library of gaussians:

    call getConvolverWidths(convolverWidths)
    call calcGaussianLibrary(convolverWidths, gaussians, convolvers)
    deallocate(convolverWidths)

    !....................................................................
    ! Allocate and calculate the squares of the convolvers:

    allocate(convsSqurd(numConvolvers))
    do i = 1, numConvolvers
      convolverXSize = size(convolvers(i)%array, 1)
      convolverYSize = size(convolvers(i)%array, 2)
      allocate(convsSqurd(i)%array(convolverXSize, convolverYSize))
      convsSqurd(i)%array = convolvers(i)%array * convolvers(i)%array
    end do

    !....................................................................
    ! Calculate some FFTs:

    maxConvolverXSize = size(convolvers(1)%array, 1)
    maxConvolverYSize = size(convolvers(1)%array, 2)

    call calcPaddedSizes(maxConvolverXSize, maxConvolverYSize&
      , localXSize, localYSize, paddedXSize, paddedYSize)

    call prepareFFTPlans(paddedXsize, paddedYsize)

    allocate(inImageFFT(paddedXsize, paddedYsize)&
      , inWeightFFT(paddedXsize, paddedYsize)&
      , inVarImageFFT(paddedXsize, paddedYsize))

      call doFft(inImage,     inImageFFT)
      call doFft(inVarImage,  inVarImageFFT)
      call doFft(weightImage, inWeightFFT)

      inImageFFT    = inImageFFT    / real(paddedXsize) / real(paddedYsize)
      inVarImageFFT = inVarImageFFT / real(paddedXsize) / real(paddedYsize)
      inWeightFFT   = inWeightFFT   / real(paddedXsize) / real(paddedYsize)

    !....................................................................
    ! Calculate the index image:

    call message(AppMsg, Verbose, 'Calculating the convolver index image.')

    allocate(indexImage(localXsize, localYsize)&
      , bestSnr(localXsize, localYsize)&
      , secondBestSnr(localXsize, localYsize)&
      , smoothedImage(localXsize, localYsize)&
      , smoothedImageVariance(localXsize, localYsize)&
      , smoothedImageUncerts(localXsize, localYsize)&
      , failureMask(localXsize, localYsize))

    where(outMask)
      indexImage = 1
    elsewhere
      indexImage = 0
    end where

    smoothedImage = 0.0
    smoothedImageVariance = 0.0
    failureMask = .false.

    !....................................................................
    ! 1st (broadest) gaussian:

    i = 1
    numPixels = count(indexImage == i)
    gaussians(i)%numPixels = numPixels
    sigma = gaussians(i)%sigma

    write(errstr, '(a, i3.3, a, f6.2, a)') '  ', i, 'th convolver, &
      &width = ',sigma, ' pixels'
    call message(AppMsg, Verbose)

if (testFlag) print *, 'calcConvolversAdaptive: inImage(txi,tyi): '&
, inImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, 'calcConvolversAdaptive: inVarImage(txi,tyi): '&
, inVarImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, 'calcConvolversAdaptive: weightImage(txi,tyi): '&
, weightImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, ''

if (testFlag) print *, 'calcConvolversAdaptive: smoothedImage(txi,tyi): '&
, smoothedImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, 'calcConvolversAdaptive: smoothedImageVariance(txi,tyi): '&
, smoothedImageVariance(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, 'calcConvolversAdaptive: failureMask(txi,tyi): '&
, failureMask(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, ''


    call maskedSmooth(convolvers(i)%array&
      , weightImage, inWeightFFT, logical(indexImage == i, bool)&
      , inImage, inImageFFT, inVarImage, inVarImageFFT&
      , smoothedImage, smoothedImageVariance, failureMask)

    ! smoothedImage = sum(conv_ij * inImage_ij) / sum(conv_ij).

    ! At present rmsSmoothedImage = sum(conv_ij^2 * var_ij) / sum(conv_ij^2).
    ! But what we want is the uncertainty of the smoothed image,
    ! which = sqrt(sum(conv_ij^2 * var_ij)) / sum(conv_ij). Hence we must
    ! do rmsSmoothedImage = sqrt(rmsSmoothedImage * sum(conv_ij^2))
    ! / sum(conv_ij).

if (testFlag) then
  print *, 'conv index =', i 
  print *, 'indexImage =', indexImage(1+txi-xStart,1+tyi-yStart) 
  print *, 'conv sigma =', sigma 
  print *, 'smoothedImage =', smoothedImage(1+txi-xStart,1+tyi-yStart) 
  print *, 'smoothedImageVariance ='&
    , smoothedImageVariance(1+txi-xStart,1+tyi-yStart) 
  print *, 'smoothedImageUncert ='&
    , sqrt(smoothedImageVariance(1+txi-xStart,1+tyi-yStart)) 
  print *, 'bestSnr ='&
    , smoothedImage(1+txi-xStart,1+tyi-yStart)&
    /sqrt(smoothedImageVariance(1+txi-xStart,1+tyi-yStart)) 
  print *, '' 
end if

    !....................................................................
    ! Decide whether to increment index:

    where(indexImage == i)
      where(.not. failureMask) ! Note that pixels for which failureMask is
      ! set remain at the present value of index. In the main smoothing
      ! routine the attempt will again be made to smooth them with the
      ! associated gaussian, which will again fail. The global failure mask
      ! will be set at this time. We might possibly get a small speed gain
      ! by attempting to do it here but at large loss of clarity and
      ! generality.

        where(smoothedImageVariance > 0.0 .and. smoothedImage > 0.0) ! This
        ! criterion needs some explanation. For smoothedImageVariance to
        ! be <= zero would be a serious bug - because varianceImage,
        ! its unsmoothed precursor, should be everywhere > 0. Hence under
        ! normal circumatances we would expect the present criterion to
        ! fail when smoothedImage <= 0. Although the input image is intended
        ! to represent 'reality' in that the parent image it is presumed to be
        ! a sample of should be everywhere positive, it may in practice be
        ! desired to smooth locally-negative input images, created for
        ! example by subtracting a uniform background from a precursor
        ! image which was formed by accumulation of Poissonian counts. Such
        ! images may easily have patches where there are, by chance, too
        ! few positive 'counts' to overcome the subtracted background level.
        ! Smoothing of such patches may generate a negative result. What to
        ! do with such pixels?
        !
        ! Recall that the index image is formed by trying first the broadest
        ! possible smoothing gaussian. If this produces a negative smoothed
        ! value for any pixel, it seems unlikely that a narrower smoother
        ! would perform otherwise. In this case probably the best we can do
        ! at this pixel is to leave the width of the smoother (designated by
        ! the value of the index image) at its present value.
        !
        ! For pixels which obey the criterion, let us proceed:

          smoothedImageUncerts = sqrt(smoothedImageVariance)
          bestSnr = smoothedImage / smoothedImageUncerts

          where(bestSnr > desiredSnr) ! This means that we are still
          ! smoothing too hard, and need to try this pixel again with a
          ! narrower gaussian. Therefore increment the index:
 
            indexImage = i + 1
          end where ! bestSnr > desiredSnr
        end where ! rmsSmoothedImage > 0.0 .and. smoothedImage > 0.0
      end where ! .not. failureMask
    end where ! indexImage == i

    !....................................................................
    ! Now all the other gaussians:

    do i = 2, numConvolvers
      ! It is perhaps worth noting that at this stage, that for any pixel
      ! (xi,yi) for which indexImage == i, the logic of the algorith so far
      ! ensures that the following are also true:
      ! 
      !   (i) outMask(xi,yi) = .true.;
      !   (ii) Smoothing has never produced a negative result at this pixel;
      !   (iii) No previous smoother has given a SNR <= desiredSnr;
      !   (iv) No previous smoother has led to failure due to insufficient
      ! >0 weightImage pixels within reach of the convolver.

      secondBestSnr = bestSnr
      numPixels = count(indexImage == i)
      gaussians(i)%numPixels = numPixels
      sigma = gaussians(i)%sigma

      write(errstr, '(a, i3.3, a, f6.2, a)') '  ', i, 'th convolver, &
        &width = ',sigma, ' pixels'
      call message(AppMsg, Verbose)

      call maskedSmooth(convolvers(i)%array&
       , weightImage, inWeightFFT, logical(indexImage == i, bool)&
       , inImage, inImageFFT, inVarImage, inVarImageFFT&
       , smoothedImage, smoothedImageVariance, failureMask)

if (testFlag .and. i == tci) then
  call dumpImageToFits(smoothedImage, 'smoothedImage.ds')
  call dumpImageToFits(smoothedImageVariance, 'smoothedImageVariance.ds')
end if

       ! smoothedImage = sum(conv_ij * inImage_ij) / sum(conv_ij).

!      ! At present rmsSmoothedImage = sum(conv_ij^2 * var_ij)
!      ! / sum(conv_ij^2). But what we want is the uncertainty of the
!      ! smoothed image, which = sqrt(sum(conv_ij^2 * var_ij))
!      ! / sum(conv_ij). Hence we must do rmsSmoothedImage
!      ! = sqrt(rmsSmoothedImage * sum(conv_ij^2)) / sum(conv_ij).

!      norm = sum(convolvers(i)%array)
!      normCorrector = sum(convsSqurd(i)%array) / norm / norm

if (testFlag) then
  print *, 'conv index =', i 
  print *, 'indexImage =', indexImage(1+txi-xStart,1+tyi-yStart) 
  print *, 'conv sigma =', sigma 
  print *, 'smoothedImage =', smoothedImage(1+txi-xStart,1+tyi-yStart) 
  print *, 'smoothedImageVariance ='&
    , smoothedImageVariance(1+txi-xStart,1+tyi-yStart) 
  print *, 'smoothedImageUncert ='&
    , sqrt(smoothedImageVariance(1+txi-xStart,1+tyi-yStart)) 
  print *, 'secondBestSnr ='&
    , secondBestSnr(1+txi-xStart,1+tyi-yStart) 
  print *, 'bestSnr ='&
    , smoothedImage(1+txi-xStart,1+tyi-yStart)&
    /sqrt(smoothedImageVariance(1+txi-xStart,1+tyi-yStart)) 
  print *, '' 
end if

      !..................................................................
      ! Decide whether to increment index:

      where(indexImage .eq. i)
        where(failureMask)
          indexImage = i - 1 ! This pixel didn't fail for the previous
          ! value of index (=> next larger gaussian), otherwise it would
          ! never have got here. Hence we will drop the index value for
          ! this pixel back to this last known 'good' value.
!*** probably ought to indicate this somewhere. It is not an absolute failure, on the other hand the SNR will not approach the desired value.

        elsewhere
          where(smoothedImageVariance > 0.0 .and. smoothedImage > 0.0)
            smoothedImageUncerts = sqrt(smoothedImageVariance)
            bestSnr = smoothedImage / smoothedImageUncerts
          elsewhere
            bestSnr = 0.0
          end where

          ! See description above for circumstances in which the above
          ! criterion can fail.
          !
          ! Why set bestSnr = 0.0 in this case? This is used to signal
          ! to the choice below. Such pixels will be tested against
          ! secondBestSnr (associated with index = i - 1), which must be
          ! > 0 (otherwise the pixel would never have got here). Either
          ! 0 (=bestSnr) < secondBestSnr < 2*desiredSnr in which case it
          ! is best to go back to secondBestSnr (hence to i-1), or
          ! secondBestSnr >= 2*desiredSnr, in which case secondBestSnr
          ! probably represents far too hard a smoothing, in which case
          ! the present value of i may be the best choice.

          where(bestSnr > desiredSnr) ! This means that we are still
          ! smoothing too hard, and need to try this pixel again with a
          ! narrower gaussian. Therefore increment the index:

            indexImage = i + 1 ! Note that this will overflow the permitted
            ! range of index if i = numConvolvers. However such overflows
            ! are truncated after the end of the loop.

          elsewhere ! go to the value that gave the best SNR:
            where(abs(secondBestSnr - desiredSnr) < abs(bestSnr - desiredSnr))
              indexImage = i - 1
              ! Otherwise, leave indexImage at i.
            end where
          end where
        end where
      end where
    end do ! gaussian number

    where(indexImage > numConvolvers) indexImage = numConvolvers

    deallocate(smoothedImageVariance, smoothedImageUncerts&
      , secondBestSnr, bestSnr&
      , inImageFFT, inVarImageFFT, inWeightFFT)

    allocate(indexImageArg(localXsize, localYsize))
    indexImageArg = indexImage

  end subroutine calcConvolversAdaptive

  !---------------------------------------------------------------------
  function divideImages(numImage, denomImage, nullValue) result(resultImage)
  !---------------------------------------------------------------------
  ! Divides numImage by denomImage. If the denominator = 0, leaves nullValue
  ! in the result.

    implicit none

    real(single), intent(in)           :: numImage(:,:),&
                                          denomImage(&
                                            size(numImage,1),size(numImage,2))
    real(single), intent(in), optional :: nullValue
    real(single)                       :: resultImage(&
                                            size(numImage,1),size(numImage,2))

    real(single) :: localNullValue

    !-------------------------------------------------------------------
    ! Start:

    if (present(nullValue)) then
      localNullValue = nullValue
    else
      localNullValue = 0.0
    end if

    where(denomImage .ne. 0.0)
      resultImage = numImage / denomImage
    elsewhere
      resultImage = localNullValue
    end where

  end function divideImages

  !---------------------------------------------------------------------
  subroutine getConvolverWidths(convolverWidths)
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: convolverWidths(:)

    !-------------------------------------------------------------------
    ! Start:

    if (withUserWidths) then
      allocate(convolverWidths(numConvolvers))
      convolverWidths = userConvolverWidths
      minConvolverWidth = minval(convolverWidths)
      maxConvolverWidth = maxval(convolverWidths)

    else
      select case(widthDistribStyle)
      case('linear')
        call calcWidthsLinearDist(convolverWidths)
      case('log')
        call calcWidthsLogDist(convolverWidths)
      case('sqrt')
        call calcWidthsSqrtDist(convolverWidths)
      case default
call error('badWidthScaleStyle',"The value of parameter --widthliststyle &
&was not recognized.")
      end select
    end if

  end subroutine getConvolverWidths

  !---------------------------------------------------------------------
  subroutine calcGaussianLibrary(convolverWidths, gaussians, convolvers)
  !---------------------------------------------------------------------
  ! The structure 'gaussian' is used to store five quantities for each of the
  ! gaussian convolvers in the library: the width or sigma value of the
  ! gaussian, the normalisation parameter (this is the sum of the elements
  ! of the unnormalised gaussian convolver array), the improvement in SNR
  ! associated with this gaussian, the half-width of the truncation boundary
  ! in pixels and the number of pixels assigned to be smoothed by this
  ! gaussian. Only the first three are calculated here: the fourth value
  ! can vary because it depends on the number of iterations applied to the
  ! image.
  !
  ! The gaussians in the library are ordered so that the lower the value
  ! of sigmaIndex, the broader the gaussian.
  !
  ! Image pixels for which the desired SNR improvement is greater than a
  ! given maximum are smoothed by the broadest gaussian in the library;
  ! those for which the desired improvement in SNR is less than or equal
  ! to 1 are smoothed by a zero-width gaussian (ie, not smoothed at all).

    implicit none

    real(single), intent(in) :: convolverWidths(:)
    type(gaussData), pointer :: gaussians(:)
    type(convolverType), pointer :: convolvers(:)

    integer(int16) ::&
      i, halfWidthPixels, convolverXSize, convolverYSize
    real(single) ::&
      sigma

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Calculating widths of the library &
    &of gaussians.')

if (numConvolvers < 1) call error('noConvolvers',"This error probably &
&means that you haven't supplied any convolvers in the list --userwidths.")
    if (numConvolvers < 2) call warning('onlyOneConvolver', 'This will &
      &work ok, but you may as well have done simple smoothing.')

if (any(convolverWidths < 0.0)) call error('convolverWidthLessThanZero'&
,"I can't think how this might occur. If you see this error, please &
&contact the code developer.")
    if (any(convolverWidths(:numConvolvers-1) <= convolverWidths(2:)))&
call error('convolverWidthsBadOrder',"This error probably means that &
&the list of convolvers you supplied in --userwidths is not in increasing &
&order.")

    allocate(gaussians(numConvolvers))
    gaussians%numPixels = 0 ! vector op, default
    allocate(convolvers(numConvolvers))

    do i = 1, numConvolvers
      sigma = convolverWidths(i)
      halfWidthPixels = convolverArrayHalfWidth(sigma)
      convolverXSize = 3 + 2 * halfWidthPixels
      convolverYSize = 3 + 2 * halfWidthPixels
      allocate(convolvers(i)%array(convolverXSize, convolverYSize))
      call calcGaussianConvolver(sigma, convolvers(i)%array)

      gaussians(i)%sigma           = sigma
      gaussians(i)%halfWidthPixels = halfWidthPixels
      gaussians(i)%norm            = sum(convolvers(i)%array)
      gaussians(i)%snrImprovement  = gaussians(i)%norm&
        / rootSumSquares(convolvers(i)%array)

      ! Notes: (i) The improvement in SNR given by the above formula assumes
      ! that the average value of the function or image to be smoothed is
      ! constant. It is the departure of real images from this ideal that
      ! requires the iteration implemented in this software. (ii) For large
      ! sigma the SNR improvement wrought by a sampled gaussian is
      ! proportional to 2 * sigma * RootPi. As  sigma -> 0 the SNR
      ! improvement departs from this rule and asymtotes to 1.
    end do

  end subroutine calcGaussianLibrary

  !---------------------------------------------------------------------
  subroutine calcWidthsLinearDist(convolverWidths)
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: convolverWidths(:)

    integer(int16) ::&
      i

    !-------------------------------------------------------------------
    ! Start:

    allocate(convolverWidths(numConvolvers))

    do i = 1, numConvolvers
      convolverWidths(i) = fracVal(maxConvolverWidth, minConvolverWidth&
        , (i - 1) / real(numConvolvers - 1))
      ! Note that the lower the convolver, the broader.
    end do

  end subroutine calcWidthsLinearDist

  !---------------------------------------------------------------------
  subroutine calcWidthsLogDist(convolverWidths)
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: convolverWidths(:)

    integer(int16) ::&
      i

    !-------------------------------------------------------------------
    ! Start:

    allocate(convolverWidths(numConvolvers))

if (minConvolverWidth < 0.0) call error('convolverWidthLessThanZero'&
,"I can't think how this might occur. If you see this error, please &
&contact the code developer.")

if (minConvolverWidth == 0.0) call error('minConvolverWidthEqualsZero'&
,"This won't work with --widthliststyle=log - you must specify a &
&--minwidth greater than zero.")

!**** this splitting of 1 error into 2 is not yet in the docs.

    do i = 1, numConvolvers
      convolverWidths(i) = fracVal(log(maxConvolverWidth)&
        , log(minConvolverWidth)&
        , (i - 1) / real(numConvolvers - 1))
      ! Note that the lower the convolver, the broader.
    end do

    convolverWidths = exp(convolverWidths)

  end subroutine calcWidthsLogDist

  !---------------------------------------------------------------------
  subroutine calcWidthsSqrtDist(convolverWidths)
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: convolverWidths(:)

    integer(int16) ::&
      i

    !-------------------------------------------------------------------
    ! Start:

    allocate(convolverWidths(numConvolvers))

    do i = 1, numConvolvers
      convolverWidths(i) = fracVal(sqrt(maxConvolverWidth)&
        , sqrt(minConvolverWidth)&
        , (i - 1) / real(numConvolvers - 1))
      ! Note that the lower the convolver, the broader.
    end do

    convolverWidths = convolverWidths * convolverWidths

  end subroutine calcWidthsSqrtDist

  !---------------------------------------------------------------------
  subroutine calcPaddedSizes(convolverXSize, convolverYSize&
  , substrateXSize, substrateYSize, paddedXSize, paddedYSize)
  !---------------------------------------------------------------------
 
    implicit none

    integer(int16), intent(in) ::&
      convolverXSize, convolverYSize, substrateXSize, substrateYSize
    integer(int16), intent(out) ::&
      paddedXSize,&
      paddedYSize

    !-------------------------------------------------------------------
    ! Start:

    ! Determine size of pad needed, then round this to nearest multiple
    ! of 2, 3, 5 & 7:

    paddedXSize = substrateXSize + convolverXSize
    paddedXSize = findNextHighest2357multiple(int(paddedXSize)) ! fftw_aux_mod
    paddedYSize = substrateYSize + convolverYSize
    paddedYSize = findNextHighest2357multiple(int(paddedYSize)) ! fftw_aux_mod

  end subroutine calcPaddedSizes

  !---------------------------------------------------------------------
  subroutine prepareFFTPlans(paddedXsize, paddedYsize)
  !---------------------------------------------------------------------
 
    implicit none

    integer(int16), intent(in) :: paddedXsize, paddedYsize

    !-------------------------------------------------------------------
    ! Start:

    if (.not. forwardPlanExists) then
      call fftw2d_f77_create_plan(fftForwardPlanNum, int(paddedXsize)&
      , int(paddedYsize), FFTW_FORWARD, FFTW_ESTIMATE + FFTW_IN_PLACE)

      forwardPlanExists = .true.
    end if

    if (.not. reversePlanExists) then
      call fftw2d_f77_create_plan(fftReversePlanNum, int(paddedXsize)&
      , int(paddedYsize), FFTW_BACKWARD, FFTW_ESTIMATE + FFTW_IN_PLACE)

      reversePlanExists = .true.
    end if

  end subroutine prepareFFTPlans

  !---------------------------------------------------------------------
  subroutine doFft(inImage, inImageFFT)
  !---------------------------------------------------------------------
 
    implicit none

    real(single),    intent(in)  :: inImage(:,:)
    complex(single), intent(out) :: inImageFFT(:,:)

    integer(int16) ::&
      imageXsize,&
      imageYsize

    !-------------------------------------------------------------------
    ! Start:

    imageXsize = size(inImage, 1)
    imageYsize = size(inImage, 2)

    inImageFFT = (0.0, 0.0) ! array op, default
    inImageFFT(1:imageXsize, 1:imageYsize) = cmplx(inImage, 0.0, single)
    call fftwnd_f77_one(fftForwardPlanNum, inImageFFT, 0)

  end subroutine doFft

  !---------------------------------------------------------------------
  subroutine doSmoothing(inImage, inVarImage, weightImage, outMask, indexImage&
  , convolvers, outImage, outVarImage, failureMask)
  !---------------------------------------------------------------------

    implicit none

    real(single),        intent(in)  :: inImage(:,:),&
                                        inVarImage(size(inImage,1)&
                                                  ,size(inImage,2)),&
                                        weightImage(size(inImage,1)&
                                                   ,size(inImage,2))
    logical(bool),       intent(in)  :: outMask(size(inImage,1)&
                                               ,size(inImage,2))
    integer(int32),      intent(in)  :: indexImage(size(inImage,1)&
                                                  ,size(inImage,2))
    type(convolverType), intent(in)  :: convolvers(:)
    real(single),        intent(out) :: outImage(size(inImage,1)&
                                                     ,size(inImage,2)),&
                                        outVarImage(size(inImage,1)&
                                                     ,size(inImage,2))
    logical(bool),       intent(out) :: failureMask(size(inImage,1)&
                                                   ,size(inImage,2))

    integer(int32) ::&
      numMaskedPixels,&
      numConvolvers, i,&
      numPixels, numDirectOps
    integer(int16) ::&
      localXSize, localYSize,&
      convolverXSize, convolverYSize,&
      maxConvolverXsize, maxConvolverYsize,&
      paddedXSize,&
      paddedYSize
    real(single) ::&
      maxLnNumDirectOps, lnNumDirectOps, mostNegValue
!,&
!      weightedInImage(size(inImage,1),size(inImage,2))
    logical ::&
      overflow
    complex(single), dimension(:,:), allocatable ::&
      inImageFFT,&
      inWeightFFT,&
      inVarImageFFT

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Now convolving the image.')

    numMaskedPixels = count(outMask)

    localXSize = size(inImage, 1)
    localYSize = size(inImage, 2)

    numConvolvers = size(convolvers)
if (maxval(indexImage) > numConvolvers) call error('badIndexImage'&
,'The maximum index found in the index image exceeds the number of convolvers.')

    !....................................................................

    outImage = 0.0
    outVarImage = 0.0
    where(outMask)
      outImage = inImage
      outVarImage = inVarImage ! default, leaves
      ! areas where (outMask .and. indexImage == 0) unsmoothed.
    end where

    ! Get maximum numOps and, if bigger than spec, make ffts of input image
    ! and inmask:
!**** ought to have some way to access these FFTs if they already exist.

    maxConvolverXSize = size(convolvers(1)%array, 1)
    maxConvolverYSize = size(convolvers(1)%array, 2)

    maxLnNumDirectOps = 0.0
    do i = 1, numConvolvers
    numPixels = count(outMask .and. indexImage == i)

      convolverXSize = size(convolvers(i)%array, 1)
      convolverYSize = size(convolvers(i)%array, 2)
      call calcNumDirectOps(convolverXSize, convolverYSize&
        , numPixels, numDirectOps, lnNumDirectOps, overflow)
    if (numDirectOps <= 0 .and. .not. overflow) cycle

      maxLnNumDirectOps = max(maxLnNumDirectOps, lnNumDirectOps)
    end do

    if ((forceCalcType .and. doCalcByFFT)&
    .or. (maxLnNumDirectOps > log(limitOnNumDirectOps)&
    .and. .not. (forceCalcType .and. .not. doCalcByFFT))) then

      call calcPaddedSizes(maxConvolverXSize, maxConvolverYSize&
        , localXSize, localYSize, paddedXSize, paddedYSize)

      call prepareFFTPlans(paddedXsize, paddedYsize)

      allocate(inImageFFT(paddedXsize, paddedYsize)&
        , inVarImageFFT(paddedXsize, paddedYsize)&
        , inWeightFFT(paddedXsize, paddedYsize))

      call doFft(inImage,     inImageFFT)
      call doFft(inVarImage,  inVarImageFFT)
      call doFft(weightImage, inWeightFFT)

      inImageFFT    = inImageFFT    / real(paddedXsize) / real(paddedYsize)
      inVarImageFFT = inVarImageFFT / real(paddedXsize) / real(paddedYsize)
      inWeightFFT   = inWeightFFT   / real(paddedXsize) / real(paddedYsize)
    end if

    ! Now do the convolutions:

    failureMask = .false. ! default
    do i = 1, numConvolvers
      numPixels = count(outMask .and. indexImage == i)

      convolverXSize = size(convolvers(i)%array, 1)
      convolverYSize = size(convolvers(i)%array, 2)
      call calcNumDirectOps(convolverXSize, convolverYSize&
        , numPixels, numDirectOps, lnNumDirectOps, overflow)

    if (numDirectOps <= 0 .and. .not. overflow) cycle

      if ((forceCalcType .and. doCalcByFFT)&
      .or. (lnNumDirectOps > log(limitOnNumDirectOps)&
      .and. .not. (forceCalcType .and. .not. doCalcByFFT))) then

        write(errstr, '(a, i3.3, a, f7.3, a)') '  ', i&
        , 'th convolver; percent pixels = ', numPixels&
        / real(numMaskedPixels) / 0.01, ' %: FFT.'

        call message(AppMsg, Verbose)

        call convolveByFft(convolvers(i)%array, inImageFFT, inVarImageFFT&
          , inWeightFFT&
          , (outMask .and. logical(indexImage == i, bool)), outImage&
          , outVarImage, failureMask)

      else
        write(errstr, '(a, i3.3, a, f7.3, a)') '  ', i&
        , 'th convolver; percent pixels = ', numPixels&
        / real(numMaskedPixels) / 0.01, ' %: Direct.'

        call message(AppMsg, Verbose)

        call convolveByDirect(convolvers(i)%array, inImage, inVarImage&
          , weightImage&
          , (outMask .and. logical(indexImage == i, bool)), outImage&
          , outVarImage, failureMask)
      end if
    end do

    ! Remove any negative values that don't belong there:
    if (all(inImage >= 0.0) .and. all(weightImage >= 0.0)&
    .and. noNegConvValues(convolvers)) then
      if (any(outImage < 0.0)) then
        mostNegValue = minval(outImage)

        write(errstr, '(a, es11.4, a)') "the most negative of which is "&
         , mostNegValue, ". The task sets all such pixels to zero. &
         &Occurrence of slightly negative values is not necessarily cause &
         &for alarm, particularly if the FFT was used at all. However you &
         &probably ought to check the output to make sure that the &
         &remaining non-negative values look ok."

        call warning('negativeValuesInOutput')

        where(outImage < 0.0) outImage = 0.0
      end if

      if (writeVarianceSet .and. any(outVarImage < 0.0)) then
        mostNegValue = minval(outVarImage)

        write(errstr, '(a, es11.4, a)') "the most negative of which is "&
         , mostNegValue, ". The task sets all such pixels to zero. &
         &Occurrence of slightly negative values is not necessarily cause &
         &for alarm, particularly if the FFT was used at all. However you &
         &probably ought to check the output variance image to make sure &
         &that the remaining non-negative values look ok."

        call warning('negativeValuesInVarianceOutput')

        where(outVarImage < 0.0) outVarImage = 0.0
      end if
    end if

  end subroutine doSmoothing

  !---------------------------------------------------------------------
  subroutine calcNumDirectOps(convolverXSize, convolverYSize&
  , numPixelsInMask, numDirectOps, lnNumDirectOps, overflow)
  !---------------------------------------------------------------------
  ! Calculates the number of multiplications that would be needed to directly
  ! convolve the portion of the image indicated by mask with the gaussian
  ! convolver of given width sigma.

    implicit none

    integer(int16), intent(in) ::&
      convolverXSize, convolverYSize
    integer, intent(in) ::&
      numPixelsInMask
    integer(int32), intent(out) ::&
      numDirectOps
    real(single), intent(out) ::&
      lnNumDirectOps
    logical, intent(out) ::&
      overflow

    !-------------------------------------------------------------------
    ! Start:

    lnNumDirectOps = log(real(numPixelsInMask))&
      + log(real(convolverXSize)) + log(real(convolverYSize))

    if (lnNumDirectOps > log(0.99 * huge(numDirectOps))) then
    ! The 0.99 converts the argument of the log() function to real as is
    ! required, but 0.99 rather than eq 1.0 also allows a bit of leeway,
    ! to make sure that no number near to huge(numDirectOps) can
    ! nonetheless slip through the gap during the conversion to real.

      overflow = .true.
      numDirectOps = -1
    else
      overflow = .false.
      numDirectOps = numPixelsInMask * convolverXSize * convolverYSize
    end if

  end subroutine calcNumDirectOps

  !---------------------------------------------------------------------
  subroutine convolveByFft(convolver, inImageFFT, inVarImageFFT, weightImageFFT&
  , outMask, outImage, outVarImage, failureMask)
  !---------------------------------------------------------------------
  ! Convolve by the Fourier method.
  !
  ! Note that outImage is declared intent(inout) rather than intent(out)
  ! since it is desirable not to mess with the values in pixels not
  ! covered by the outmask. These may already have been smoothed using
  ! other convolver kernels and should therefore be retained.

    implicit none

    real(single),     intent(in)    :: convolver(:,:)
    complex(single),  intent(in)    :: inImageFFT(:,:),&
                                       inVarImageFFT(&
                                         size(inImageFFT,1),&
                                         size(inImageFFT,2)),&
                                       weightImageFFT(&
                                         size(inImageFFT,1),&
                                         size(inImageFFT,2))
    logical(bool),    intent(in)    :: outMask(:,:)
    real(single),     intent(inout) :: outImage(&
                                         size(outMask,1),size(outMask,2)),&
                                       outVarImage(&
                                         size(outMask,1),size(outMask,2))
    logical(bool),    intent(inout) :: failureMask(&
                                         size(outMask,1),size(outMask,2))

    real(single), dimension(size(outMask, 1), size(outMask, 2)) ::&
      tempInImage,&
      tempVarImage,&
      weightImage
    complex(single), dimension(size(inImageFFT, 1), size(inImageFFT, 2)) ::&
      tempInImageFFT,&
      tempVarImageFFT,&
      tempWeightImageFFT,&
      convolverFFT,&
      convolverSquaredFFT
    integer(int16) ::&
      localXsize, localYsize, paddedXsize, paddedYsize,&
      convolverXSize, convolverYSize, halfConvolverXSize, halfConvolverYSize
    real(single) :: convolverNorm, normalizer, localMinAllowedWeight

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, 'start of convolveByFft'

  if (.not. any(outMask)) return ! outmask is empty.

if (testFlag) print *, '  convolveByFft: xStart, yStart, txi-xStart+1, tyi-yStart+1: '&
, xStart, yStart, txi-xStart+1,tyi-yStart+1
if (testFlag) print *, ''

if (testFlag) print *, '  convolveByFft: inImageFFT(txi,tyi): '&
, inImageFFT(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByFft: inVarImageFFT(txi,tyi): '&
, inVarImageFFT(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByFft: weightImageFFT(txi,tyi): '&
, weightImageFFT(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByFft: outMask(txi,tyi): '&
, outMask(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, ''

if (testFlag) print *, '  convolveByFft: outImage(txi,tyi): '&
, outImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByFft: outVarImage(txi,tyi): '&
, outVarImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByFft: failureMask(txi,tyi): '&
, failureMask(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, ''

    paddedXsize = size(inImageFFT, 1)
    paddedYsize = size(inImageFFT, 2)
    localXsize = size(outMask, 1)
    localYsize = size(outMask, 2)
    convolverXSize = size(convolver, 1)
    convolverYSize = size(convolver, 2)
    halfConvolverXSize = floor(convolverXSize / 2.0)
    halfConvolverYSize = floor(convolverYSize / 2.0)

    convolverFFT = (0.0, 0.0) ! array op, default
    convolverFFT(1:convolverXSize, 1:convolverYSize)&
      = cmplx(convolver, 0.0, single)
    call recenter(convolverFFT, int(halfConvolverXSize) + 1&
      , int(halfConvolverYSize) + 1) ! in ssclib/fftw_aux
!*** couldn't you use the fortran intrinsic function cshift instead of recenter?
    call fftwnd_f77_one(fftForwardPlanNum, convolverFFT, 0)

    convolverSquaredFFT = (0.0, 0.0) ! array op, default
    convolverSquaredFFT(1:convolverXSize, 1:convolverYSize)&
      = cmplx(convolver*convolver, 0.0, single)
    call recenter(convolverSquaredFFT, int(halfConvolverXSize) + 1&
      , int(halfConvolverYSize) + 1) ! in ssclib/fftw_aux
!*** couldn't you use the fortran intrinsic function cshift instead of recenter?
    call fftwnd_f77_one(fftForwardPlanNum, convolverSquaredFFT, 0)

    convolverNorm = sum(convolver)
!    localMinAllowedWeight = minAllowedWeight * convolverNorm&
!      * paddedXsize * paddedXsize
localMinAllowedWeight = minAllowedWeight * convolverNorm
if (testFlag) print *, '  convolveByFft: localMinAllowedWeight: '&
, localMinAllowedWeight
if (testFlag) print *, ''

    if (normalizePlease) then
      normalizer = 1.0
    else
      normalizer = convolverNorm
      if (normalizer == 0.0)& ! note that <0 is permitted.
call error('zeroNormConvolver',"The sum of the convolver elements is zero. &
&Such convolvers are not allowed.")
    end if

    tempInImageFFT = inImageFFT * convolverFFT
    call fftwnd_f77_one(fftReversePlanNum, tempInImageFFT, 0)
    tempInImage = real(tempInImageFFT(1:localXsize,1:localYsize), single)

    tempWeightImageFFT = weightImageFFT * convolverFFT
    call fftwnd_f77_one(fftReversePlanNum, tempWeightImageFFT, 0)
    weightImage = real(tempWeightImageFFT(1:localXsize,1:localYsize), single)

    tempVarImageFFT = inVarImageFFT * convolverSquaredFFT
    call fftwnd_f77_one(fftReversePlanNum, tempVarImageFFT, 0)
    tempVarImage = real(tempVarImageFFT(1:localXsize,1:localYsize), single)

if (testFlag) print *, '  convolveByFft: tempInImage(txi,tyi): '&
, tempInImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByFft: weightImage(txi,tyi): '&
, weightImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByFft: tempVarImage(txi,tyi): '&
, tempVarImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, ''

    where (outMask)
      where(weightImage >= localMinAllowedWeight)
        outImage = tempInImage * normalizer / weightImage
        outVarImage = tempVarImage * normalizer * normalizer&
          / weightImage / weightImage
      elsewhere
        outImage = 0.0
        outVarImage = 0.0
        failureMask = .true. ! indicates those pixels where the weight
        ! is too small.
      end where
    end where

!if (testFlag) call dumpImageToFits(weightImage, 'dump_weight.ds')

if (testFlag) print *, '  convolveByFft: outImage(txi,tyi): '&
, outImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByFft: outVarImage(txi,tyi): '&
, outVarImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByFft: failureMask(txi,tyi): '&
, failureMask(1+txi-xStart,1+tyi-yStart)

if (testFlag) print *, 'end of convolveByFft'

  end subroutine convolveByFft

  !---------------------------------------------------------------------
  subroutine convolveByDirect(convolver, inImage, inVarImage, weightImage&
  , outMask, outImage, outVarImage, failureMask)
  !---------------------------------------------------------------------
  ! Convolve by direct (ie non-Fourier) convolution.
  !
  ! Note that outImage is declared intent(inout) rather than intent(out)
  ! since it is desirable not to mess with the values in pixels not
  ! covered by the outmask. These may already have been smoothed using
  ! other convolver kernels and should therefor be retained.

    implicit none

    real(single),  intent(in)    :: convolver(:,:),&
                                    inImage(:,:),&
                                    inVarImage(&
                                      size(inImage,1),size(inImage,2)),&
                                    weightImage(size(inImage,1),size(inImage,2))
    logical(bool), intent(in)    :: outMask(size(inImage,1),size(inImage,2))
    real(single),  intent(inout) :: outImage(size(inImage,1),size(inImage,2)),&
                                    outVarImage(size(inImage,1),size(inImage,2))
    logical(bool), intent(inout) :: failureMask(size(inImage,1)&
                                               ,size(inImage,2))

    integer(int16) ::&
      outStartX, outFinisX, outStartY, outFinisY,&
      localXsize, localYsize,&
      convolverXSize, convolverYSize, halfConvolverXSize, halfConvolverYSize,&
      xi, yi, cxi, cyi, xxi, yyi
    real(single) ::&
!      convTimesWeight,&
      convolvedWeightImage, convolvedImage, convolvedVarImage,&
      convolverNorm, normalizer, localMinAllowedWeight,&
      convSquared(size(convolver,1),size(convolver,2))

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, 'start of convolveByDirect'

  if (.not. any(outMask)) return ! outmask is empty.

if (testFlag) print *, '  convolveByDirect: xStart, yStart, txi-xStart+1, tyi-yStart+1: '&
, xStart, yStart, txi-xStart+1,tyi-yStart+1

if (testFlag) print *, '  convolveByDirect: inImage(txi,tyi): '&
, inImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByDirect: inVarImage(txi,tyi): '&
, inVarImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByDirect: weightImage(txi,tyi): '&
, weightImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByDirect: outMask(txi,tyi): '&
, outMask(1+txi-xStart,1+tyi-yStart)

if (testFlag) print *, '  convolveByDirect: outImage(txi,tyi): '&
, outImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByDirect: outVarImage(txi,tyi): '&
, outVarImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByDirect: failureMask(txi,tyi): '&
, failureMask(1+txi-xStart,1+tyi-yStart)


    localXsize = size(inImage, 1)
    localYsize = size(inImage, 2)
if (testFlag) print *, '  convolveByDirect: localXsize, localYsize: ', localXsize, localYsize
    convolverXSize = size(convolver, 1)
    convolverYSize = size(convolver, 2)
    halfConvolverXSize = floor(convolverXSize / 2.0)
    halfConvolverYSize = floor(convolverYSize / 2.0)

    call maskArrayLimits(outMask, outStartX, outFinisX, outStartY, outFinisY)

    convSquared = convolver * convolver
    convolverNorm = sum(convolver)
    localMinAllowedWeight = minAllowedWeight * convolverNorm
if (testFlag) print *, '  convolveByFft: localMinAllowedWeight: '&
, localMinAllowedWeight
if (testFlag) print *, ''

    if (normalizePlease) then
      normalizer = 1.0
    else
      normalizer = convolverNorm
      if (normalizer == 0.0)& ! note that <0 is permitted.
call error('zeroNormConvolver',"The sum of the convolver elements is zero. &
&Such convolvers are not allowed.")
    end if

    ! Now the actual convolution:
    do xi = outStartX, outFinisX
      do yi = outStartY, outFinisY
!if (testFlag .and. (xi .ne. 1+txi-xStart .or. yi .ne. 1+tyi-yStart)) cycle

      if (.not. outMask(xi, yi)) cycle

        ! The pixel (xi, yi) has been established as being within the outMask
        ! and therefore it is a valid pixel for which to calculate the
        ! convolution sum.

if (testFlag .and. xi == 1+txi-xStart .and. yi == 1+tyi-yStart) print *, ''
        convolvedWeightImage = 0.0
        convolvedImage    = 0.0
        convolvedVarImage = 0.0

        do cxi = 1, convolverXSize
          xxi = xi - (cxi - 1 - halfConvolverXSize)
        if (xxi <= 0 .or. xxi > localXsize) cycle

          do cyi = 1, convolverYSize
            yyi = yi - (cyi - 1 - halfConvolverYSize)
          if (yyi <= 0 .or. yyi > localYsize) cycle

          if (weightImage(xxi, yyi) <= 0.0) cycle

            ! The pixel (xxi, yyi) has been established as lying within
            ! a non-zero region of weightImage and is therefore available
            ! to be included in the convolution sum for pixel (xi, yi).

!            convTimesWeight = convolver(cxi, cyi) * weightImage(xxi, yyi)

            convolvedWeightImage = convolvedWeightImage + convolver(cxi, cyi)&
              * weightImage(xxi, yyi)
            convolvedImage  = convolvedImage + convolver(cxi, cyi)&
              * inImage(xxi, yyi)
            convolvedVarImage = convolvedVarImage + convSquared(cxi, cyi)&
              * inVarImage(xxi, yyi)

!if (testFlag) print *, '  convolveByDirect:   cxi, cyi,  xxi, yyi, i(xxi,yyi), !w(xxi,yyi), v(xxi,yyi), conv(cxi,cyi): '&
!, cxi, cyi, xxi, yyi, inImage(xxi, yyi), weightImage(xxi, yyi)&
!, inVarImage(xxi, yyi), convolver(cxi, cyi)
          end do
        end do
!if (testFlag) print *, ''

        if (convolvedWeightImage >= localMinAllowedWeight) then
          outImage(xi, yi) = convolvedImage * normalizer / convolvedWeightImage
          outVarImage(xi, yi) = convolvedVarImage * normalizer * normalizer&
            / convolvedWeightImage / convolvedWeightImage
        else
          outImage(xi, yi) = 0.0
          outVarImage(xi, yi) = 0.0
          failureMask(xi, yi) = .true. ! indicates those pixels where the
          ! weight is too small.
        end if

!if (testFlag) print *, '  convolveByDirect: sum, weight, varsum, outImage(txi, tyi), outVarImage(txi, tyi): '&
!, convolvedImage, convolvedWeightImage, convolvedVarImage&
!, outImage(xi, yi), outVarImage(xi, yi)

if (testFlag .and. xi == 1+txi-xStart .and. yi == 1+tyi-yStart) then
  print *, '  convolveByDirect: tempInImage(txi,tyi): '&
    , convolvedImage
  print *, '  convolveByDirect: weightImage(txi,tyi): '&
    , convolvedWeightImage
  print *, '  convolveByDirect: tempVarImage(txi,tyi): '&
    , convolvedVarImage
  print *, ''
end if

      end do
    end do

if (testFlag) print *, '  convolveByDirect: outImage(txi,tyi): '&
, outImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByDirect: outVarImage(txi,tyi): '&
, outVarImage(1+txi-xStart,1+tyi-yStart)
if (testFlag) print *, '  convolveByDirect: failureMask(txi,tyi): '&
, failureMask(1+txi-xStart,1+tyi-yStart)

if (testFlag) print *, 'end of convolveByDirect'

  end subroutine convolveByDirect

  !---------------------------------------------------------------------
  subroutine calcAndSaveTemplate(gaussians, indexImage)
  !---------------------------------------------------------------------
    
    implicit none

    type(gaussData), intent(in) ::&
      gaussians(:)
    integer(int32), intent(in) :: indexImage(:,:)

    integer(int32) :: i
    real(single), allocatable ::&
      templateImage(:,:)
    type(DataSetT) :: outSet
    type(ArrayT) :: outArray

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Writing template image '&
      //trim(templateSetName)//'.')

    allocate(templateImage(xSize, ySize))

    templateImage = 0.0 ! the minimum sensible value. This will remain the
    ! value for all those pixels for which indexImage = 0, which amounts to
    ! those for which mask = 0.

    do i = 1, size(gaussians)
      where(indexImage .eq. i)&
        templateImage(xStart:xFinis,yStart:yFinis) = gaussians(i)%sigma
    end do

    call saveImage(templateImage, inSetName, templateSetName, tempImageSetName)

    outSet = dataSet(templateSetName, MODIFY)
    outArray = array(outSet, 'PRIMARY')
    call setAttribute(outArray, "FORMAT", "2.7", "Refers to the way in &
      &which the template values are defined.")

    call release(outSet)

    deallocate(templateImage)

  end subroutine calcAndSaveTemplate

  !---------------------------------------------------------------------
  subroutine readIndexAndConvolvers(indexImage, convolvers)
  !---------------------------------------------------------------------

    implicit none

    integer(int32),      pointer :: indexImage(:,:)
    type(convolverType), pointer :: convolvers(:)

    real(single), pointer :: convCubeData(:,:,:), convArrayData(:,:)
    type(DataSetT) :: indexImageSet, convSet
    type(ArrayT) :: indexArray, convCube
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    ! Load the index image and the convolvers:

    convSet = dataSet(convolversSetName, READ)

    if (withIndexImage) then
      indexImageSet = dataSet(indexImageSetName, READ)
      indexArray = array(indexImageSet, 'INDEXIMG')
      call readArrayData(indexArray, indexImage)
      ! in ssclib/dal_aux.
      call release(indexImageSet)

      if (size(indexImage, 1) .ne. xsize .or. size(indexImage, 2) .ne. ysize)&
call error('indexImageWrongSize',"The dimensions of --inindeximageset are &
&different to those of --inset.")

      call selectArraySubset(indexImage, xStart, xFinis, yStart, yFinis)
      ! in ssclib/array_utils.

if (minval(indexImage) < 0) call error('indexImageNegativeValues',"Some of &
&the pixels of \param{inindeximageset} have negative values.")
      ! 0 is allowed, it is used to signal no smoothing.

      convCube = array(convSet, 'CONV_000')
      convCubeData => real32Array3Data(convCube)
      numConvolvers = size(convCubeData, 3)
      allocate(convolvers(numConvolvers))
      do i = 1, numConvolvers
        allocate(convolvers(i)%array(size(convCubeData, 1)&
          , size(convCubeData, 2)))
        convolvers(i)%array = convCubeData(:,:,i-1)
      end do

    else
      allocate(indexImage(xStart:xFinis, yStart:yFinis))
      indexImage = 1

      convCube = array(convSet, 0)
      convArrayData => real32Array2Data(convCube)
      allocate(convolvers(1))
      allocate(convolvers(1)%array(size(convArrayData, 1)&
        , size(convArrayData, 2)))
      convolvers(1)%array = convArrayData
    end if

    call release(convSet)

  end subroutine readIndexAndConvolvers

  !---------------------------------------------------------------------
  subroutine writeIndexAndConvolvers(indexImage, convolvers)
  !---------------------------------------------------------------------

    implicit none

    integer(int32),      intent(in) :: indexImage(:,:)
    type(convolverType), intent(in) :: convolvers(:)

    integer(int32), allocatable :: indexImageOrigSize(:,:)
    real(single), pointer :: convCubeData(:,:,:)
    type(DataSetT) :: indexImageSet, convSet
    type(ArrayT) :: indexArray, convCube
!    integer(int32) :: i, numConvolvers
integer :: i, numConvolvers
    integer(int32), pointer ::&
      newDimVector(:)
    integer(int32), pointer ::&
      arrayDataInt32(:,:)
!    integer(int16) ::&
integer ::&
      convXSize, convYSize, maxConvolverXSize, maxConvolverYSize,&
      convStartXi, convStartYi, convFinisXi, convFinisYi

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Writing index image '&
    //trim(indexImageSetName)//'.')

    ! Load the index image and the convolvers:

    allocate(indexImageOrigSize(xsize, ysize))
    indexImageOrigSize = 0
    indexImageOrigSize(xStart:xFinis, yStart:yFinis) = indexImage

    allocate(newDimVector(0:1))
!    newDimVector = (/size(indexImage,1), size(indexImage,2)/)
newDimVector = (/size(indexImageOrigSize,1), size(indexImageOrigSize,2)/)

    indexImageSet = dataSet(indexImageSetName, CREATE)
    indexArray = addArray(indexImageSet, 'INDEXIMG', INTEGER32, newDimVector)

    arrayDataInt32 => int32Array2Data(indexArray)
!    arrayDataInt32 = indexImage
arrayDataInt32 = indexImageOrigSize
deallocate(indexImageOrigSize)

    deallocate(newDimVector)
    call release(indexImageSet)


    call message(AppMsg, Verbose, 'Writing convolver dataset '&
    //trim(convolversSetName)//'.')

    numConvolvers = size(convolvers)
!*** error if < 1?
    maxConvolverXSize = size(convolvers(1)%array,1)
    maxConvolverYSize = size(convolvers(1)%array,2)
    do i = 1, numConvolvers
      maxConvolverXSize = max(maxConvolverXSize, size(convolvers(i)%array,1))
      maxConvolverYSize = max(maxConvolverYSize, size(convolvers(i)%array,2))
    end do

    ! This palaver is necessary because the convolver arrays are not
    ! necessarily all the same size.

    allocate(newDimVector(0:2))
!    newDimVector = (/size(convolvers(1)%array,1), size(convolvers(1)%array,2)&
!      , numConvolvers/)
newDimVector = (/maxConvolverXSize, maxConvolverYSize, numConvolvers/)

    convSet = dataSet(convolversSetName, CREATE)
    convCube = addArray(convSet, 'CONV_000', REAL32, newDimVector)
    convCubeData => real32Array3Data(convCube)
    convCubeData = 0.0 ! default
    do i = 1, numConvolvers
      convXSize = size(convolvers(i)%array,1)
      convYSize = size(convolvers(i)%array,2)
      convStartXi = floor(1.75 + (maxConvolverXSize - convXSize) / 2.0) - 1
      convStartYi = floor(1.75 + (maxConvolverYSize - convYSize) / 2.0) - 1
      convFinisXi = convStartXi + convXSize - 1
      convFinisYi = convStartYi + convYSize - 1
      convCubeData(convStartXi:convFinisXi,convStartYi:convFinisYi,i-1)&
        = convolvers(i)%array

      if (normalizePlease)&
        convCubeData(:,:,i-1) = convCubeData(:,:,i-1) / sum(convolvers(i)%array)
      ! This is done because at time of writing (2004-04-21) the default
      ! for --normalizeset is 'no' (a sensible default for the general
      ! case). Hence if the adaptive smoothing was done (as it normally
      ! ought to be) with normalization turned on; the resulting set of
      ! convolvers saved to --outconvolverset; then these reused again
      ! with --smoothstyle=set but with --normalizeset left at default
      ! 'no', strange results are obtained. Normalization at the present
      ! stage avoids this situation.
    end do

    deallocate(newDimVector)
    call release(convSet)

  end subroutine writeIndexAndConvolvers

  !---------------------------------------------------------------------
  real(single) function rootSumSquares(array)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: array(:,:)

    !-------------------------------------------------------------------
    ! Start

    rootSumSquares = sqrt(sum(array * array))

  end function rootSumSquares

  !---------------------------------------------------------------------
  subroutine maskedSmooth(convolver, weightImage, inWeightFFT, outMask&
  , inImage, inImageFFT, inVarImage, inVarImageFFT, outImage, outVarImage&
  , failureMask)
  !---------------------------------------------------------------------

    implicit none

    real(single),    intent(in)    :: inImage(:,:),&
                                      inVarImage(size(inImage,1)&
                                                ,size(inImage,2)),&
                                      weightImage(size(inImage,1)&
                                                 ,size(inImage,2))
    real(single),    intent(in)    :: convolver(:,:)
    logical(bool),   intent(in)    :: outMask(size(inImage,1),size(inImage,2))
    complex(single), intent(in)    :: inImageFFT(:,:),&
                                      inVarImageFFT(size(inImageFFT,1)&
                                                   ,size(inImageFFT,2)),&
                                      inWeightFFT(size(inImageFFT,1)&
                                                 ,size(inImageFFT,2))
    real(single),    intent(out)   :: outImage(size(inImage,1)&
                                                   ,size(inImage,2)),&
                                      outVarImage(size(inImage,1)&
                                                 ,size(inImage,2))
    logical(bool),   intent(inout) :: failureMask(size(inImage,1)&
                                                 ,size(inImage,2))

    real(single) :: lnNumDirectOps
    integer(int32) ::&
      numDirectOps ! Take care that this is not overflowed. A 600x600 array
      ! with a 78x78 convolver will do so.
    logical :: overflow
    integer(int16) :: convolverXSize, convolverYSize

    !-------------------------------------------------------------------
    ! Start

    convolverXSize = size(convolver, 1)
    convolverYSize = size(convolver, 2)

    call calcNumDirectOps(convolverXSize, convolverYSize, count(outMask)&
      , numDirectOps, lnNumDirectOps, overflow)

  if (numDirectOps <= 0 .and. .not. overflow) return

    if ((forceCalcType .and. doCalcByFFT)&
    .or. (lnNumDirectOps > log(limitOnNumDirectOps)&
    .and. .not. (forceCalcType .and. .not. doCalcByFFT))) then

      call convolveByFft(convolver, inImageFFT, inVarImageFFT, inWeightFFT&
        , outMask, outImage, outVarImage, failureMask)

    else
      call convolveByDirect(convolver, inImage, inVarImage, weightImage&
        , outMask, outImage, outVarImage, failureMask)
    end if

  end subroutine maskedSmooth

  !---------------------------------------------------------------------
  logical function noNegConvValues(convolvers)
  !---------------------------------------------------------------------

    implicit none

    type(convolverType), intent(in)  :: convolvers(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    noNegConvValues = .true. ! default
    do i = 1, size(convolvers)
      if (any(convolvers(i)%array < 0.0)) then
        noNegConvValues = .false.
    exit
      end if
    end do

  end function noNegConvValues

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module asmooth_mod
